#!/usr/bin/perl -w

# Pourpre scoring script for complex questions
# Copyright (C) 2005-2007, Dina Demner-Fushman and Jimmy Lin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

$stemmer = "common/porter.pm";
require $stemmer;
use Getopt::Std;

#usage
$usage =
"pourpre-1.1c.pl: Pourpre v1.1 scoring script for complex questions (count)\
\
USAGE\
 Required arguments:\
\
     -s file\
       file containing a system run in TREC QA submission format\
       (qid tag docid answer-string-as-the-remainder-of-the-line)\
\
     -r file\
       file containing answers as provided by NIST \
       (qid nugnum vital/okay string)\
\
 Optional arguments:\
     -b int\
       beta weight for f_score (recall is beta times as important as\
       precision) default = 3\
\
     -v\
       verbose - prints nuggets found in an answer
\
     -m [micro|macro]\
       averaging method (default = macro)\
        macroaveraging - treats every question as equally important\
        microaveraging - treats every nugget as equally important\
\
     -p\
       stem terms (default: no stemming)\
\
     -x file\
       stopword list default: \"common/english.stop\"\
\
     -h\
       prints this message\
";

#initialize
use vars qw ($opt_s $opt_r $opt_b $opt_h $opt_m $opt_p $opt_x $opt_v $opt_c);
getopts ('s:r:b:hm:px:vc:');
die $usage if $opt_h;

if(defined ($opt_s)){ 
    $infile = $opt_s;
} else {
    die "system run file not defined\n usage: $0 [-h] -s system_run -r reference_file\n";
}

if(defined ($opt_r)){ 
    $inrels = $opt_r;
} else {
    die "nist answers file not defined\n usage: $0 [-h] -s system_run -r reference_file\n";
} 

if(defined ($opt_b)){ 
    $beta = $opt_b;
} else {
    $beta = 3;
}

# this is a hidden option used during development only: 

# $count_mode = 2 means that each nugget is treated either as matched
# or not matched; otherwise, each nugget can be treated as being
# partially matched (i.e., fraction of matching unigrams)

$count_mode = 1;

if(defined ($opt_m)){ 
    $mode  = $opt_m;
} else {
    $mode  = "macro";
} 
 
if(defined ($opt_x)){ 
    $stopfile  = $opt_x;
} else {
    $stopfile  = "common/english.stop";
} 

if(defined ($opt_p)){ 
    $stem = 1;
} else {
    $stem = 0;
}

if(defined ($opt_v)){ 
    $verbose = 1;
} else {
    $verbose = 0;
}

$allowance = 100;

#load stopwords
%stopwords=();

open(INFILE,$stopfile) || die "Cannot open $stopfile\n";
while ($line = <INFILE>) {
    chomp $line;
    $stopwords{$line}=1;
}
close (INFILE);


#load answer key; store idf for nugget words
%nvalues = ();
%nuggets = ();
%max_nugget_count = ();
%max_qid_count = ();
@qids = ();

open(INFILE,$inrels) || die "Cannot open $inrels\n";
while ($line = <INFILE>) {
    chomp $line;
    next if ($line =~ /^s*$/);
    if ($line =~ /^Qid\s+(\d+\.?\d*)/i){
	$qid = $1;
	$max_qid_count{$qid}=0;
	push @qids, $qid;
	next;
    }
    ($qid,$id,$nval,$nugget) = split /\s+/, $line, 4;
    $max_nugget_count{$qid}{$id} = 0;
    $nugget =~ s/\W/ /g;
    $nugget =~ s/_/ /g;
    $nugget = lc $nugget;

    @answer_words = split /\s+/,$nugget;
    foreach $word (@answer_words){ 
	unless(length($word)<3 || exists $stopwords{$word}){
	    if($stem == 1){
		$word = porter($word);
	    }
        
	    $nuggets{$qid}{$id}{$word}=1;
	    $max_nugget_count{$qid}{$id}++;
	}
    }
    $nvalues{$qid}{$id}=$nval;

    if($nvalues{$qid}{$id} eq "vital"){
	$max_qid_count{$qid}++;
    }

}

close (INFILE);

if (-e $infile){
    if (-f $infile){
	&computeFscore($infile);
    }
    if (-d $infile){
	opendir (IND, $infile) || die " failed to open $infile\n";
	foreach $insub (sort grep(!/^\.\.?$/,readdir(IND))) {
	    &computeFscore($infile."\/".$insub);
	}
	close(IND);
    }
}
else{ die "Could not find file $infile\n";}

#load submission
sub computeFscore {
    my $insub=shift;
    my %lengths = ();
    my %found_allowances = ();
    my %found_nuggets=();
    my %nugget_answer_match = ();
  
    open(INFILE,$insub) || die "couldn't open $insub\n";
    while (<INFILE>) {
	chomp;
	if( /(\d+\.?\d*)\s+([\w-]+)\s+\d*\s+\S+\s+(.+)/){
	    my $qid = $1;
	    $submission = $2;
	    my $text = $3;
	    my $textl = $text;
	    $textl =~ s/\s//g;
	    if(exists $lengths{$qid}){
		$lengths{$qid}+= length $textl;
	    } else{
		$lengths{$qid}= length $textl;
	    }
	    $text =~ s/\W/ /g;
	    $text =~ s/_/ /g;
	    $text = lc $text;
	    my @textwords = split /\s+/,$text;
	    foreach my $nugget_id (keys %{$nuggets{$qid}}){
		my $match_count = 0;
		foreach my $word (@textwords){
		    if($stem == 1){
			$word = porter($word);
		    }
		    
		    if(exists $nuggets{$qid}{$nugget_id}{$word}){  
			$match_count++;
		    }
		}
		
		if($match_count > $max_nugget_count{$qid}{$nugget_id} ){
		    $match_count = $max_nugget_count{$qid}{$nugget_id};
		}
	
		if($match_count/$max_nugget_count{$qid}{$nugget_id} > 0.005){
		    $found_allowances{$qid}{$nugget_id}=$allowance;
		    if($nvalues{$qid}{$nugget_id} eq "vital"){
			if(exists $found_nuggets{$qid}{$nugget_id}){
			    if($match_count > $found_nuggets{$qid}{$nugget_id}){
				$found_nuggets{$qid}{$nugget_id} = $match_count;
				$nugget_answer_match{$qid}{$nugget_id} = $text;
			    }
			}else{
			    $found_nuggets{$qid}{$nugget_id} = $match_count;
			    $nugget_answer_match{$qid}{$nugget_id} = $text;
			}   
		    } #end found vital nugget
		} #end found nugget
	    } #end iterating over nuggets
	} #end processing one answer
    } #end iterating over answers
    
    close INFILE;
    
    if($verbose == 1){
	foreach my $qid (keys %nugget_answer_match){
	    my %answer_nuggets = ();
	    foreach my $nugget_id (keys %{$nugget_answer_match{$qid}}){
		my $text = $nugget_answer_match{$qid}{$nugget_id};
		if(exists $answer_nuggets{$text}){
		    $answer_nuggets{$text}.=",".$nugget_id;
		} else{
		    $answer_nuggets{$text} = $nugget_id;
		}
	    }
	    foreach my $text (keys %answer_nuggets){
		print "$submission: $qid: $text\nNuggets assigned: $answer_nuggets{$text}\n";
	    }
	}
    }
    
#compute f-score
    $beta2 = $beta*$beta;
    $b1 = $beta2 + 1;
    
    if($mode eq 'micro' || $mode eq 'both'){
	my $score = 0;
	my $length = 0;
	my $allowance = 0;
	my $total_max_count = 0;
        foreach my $qid (@qids){ 
	    foreach my $nugget_id (keys %{$found_nuggets{$qid}}){	
		if($count_mode == 2){
		    $score++;
		}
		else{
		    $score += $found_nuggets{$qid}{$nugget_id};
		}
		
		$allowance += $found_allowances{$qid}{$nugget_id};		
	    }
	    $total_max_count +=$max_qid_count{$qid};
	    $length +=  $lengths{$qid};
	}
	
	if($total_max_count > 0){
	    $score /= $total_max_count;
	    $score = ($score >1) ? 1 : $score;
	}
	else{
	    $score = 0;
	}
	my $f = 0;
	my $fp = 0;
	if($score > 0){
	    if($length > 0){
		if ($length <= $allowance){
		    $fp = 1;
		}
		else{
		    $fp = 1 - ($length-$allowance)/$length;
		}
	    }
	    $f = ($b1*$score*$fp)/($beta2*$fp + $score);
	}
	if($mode eq 'both'){
	    print "$submission\tmicro-average:\t$f\n";
	}else{
	    print "$submission\t$f\n";
	}
    }
    
    if($mode eq 'macro' || $mode eq 'both'){
	my $question_count = @qids;
	my $f = 0;
	foreach my $qid (@qids){
	    my $score = 0;
	    my $allowance = 0;
	    my $length = $lengths{$qid};
	    foreach my $nugget_id (keys %{$found_nuggets{$qid}}){	 
		if($count_mode == 2){	
		    $score++;
		}
		else{
		    if( $max_nugget_count{$qid}{$nugget_id} > 0 ){
			$score += $found_nuggets{$qid}{$nugget_id}/$max_nugget_count{$qid}{$nugget_id};
		    }
		}
		
		$allowance += $found_allowances{$qid}{$nugget_id};
	    }
	    my $fp = 0;
	    $score/=$max_qid_count{$qid};
	    
	    $score = ($score >1) ? 1 : $score;
	    
	    if($score > 0 && $allowance > 0){
		if($length > 0){
		    if ($length <= $allowance){
			$fp = 1;
		    }
		    else{
			$fp = 1 - ($length-$allowance)/$length;
		    }
		}
		$f += (($b1*$score*$fp)/($beta2*$fp + $score));
	    }
	}
	$f/=$question_count;
	    
	if($mode eq 'both'){
	    print "$submission\tmacro-average:\t$f\n";
	}else{
	    print "$submission\t$f\n";
	}
    }
}

